import dataclasses
import os
import re
from collections import OrderedDict

import click
import libcst
from libcst import ClassDef, FunctionDef, IndentedBlock, Module, SimpleStatementLine

from gutt.model import Code, ModuleIO
from gutt.template import Template as T
from gutt.utils import blacking, expand_sys_path, makefile, populate_init


class InvalidModule(Exception):
    pass


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return

    from gutt import __version__

    click.echo(f"Version: {__version__}")

    ctx.exit()


@click.command()
@click.pass_context
@click.option(
    "--version",
    "-V",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
)
@click.option(
    "--modname", "-m", help="Target module name for generating test templates"
)
@click.option(
    "--path",
    "-p",
    multiple=True,
    default=[os.getcwd()],
    help='Insert into "sys.path" to search modules, could assign with multiple values',
)
@click.option(
    "--exclude",
    "-e",
    help="Giving regex pattern to match the implementation to be excluded by its qualname.",
)
@click.option(
    "--output",
    "-o",
    default="tests/_gutt",
    help='Output root directory for populating test files, default: "tests/_gutt".',
)
@click.option(
    "--template-class",
    "-t",
    default="gutt.template.Template",
    help='Assign template class to generate test template, default: "gutt.template.Template".',
)
@click.option(
    "--dryrun",
    "-dr",
    is_flag=True,
    help="Run in dryrun mode.",
)
@click.option(
    "--flatten",
    is_flag=True,
    help="Flatten the nested structure of the test module to the single folder.",
)
def main(ctx, modname, path, exclude, output, template_class, dryrun, flatten):
    head = "" if flatten else modname.split(".")[0]
    with expand_sys_path(*path):
        module: ModuleIO = ModuleIO.from_name(modname, output, head)
        Template = T.load(template_class)

        if module is None:
            raise InvalidModule(modname)

        for mod in module.iter_submodules(head):
            code_added = 0

            if isinstance(exclude, str) and re.search(exclude, mod.name):
                click.secho("ignoring module: ", nl=False, fg="bright_white")
                click.secho(mod.name, fg="bright_black")
                continue

            try:
                with open(mod.src, "r") as f:
                    src_mod = libcst.parse_module(f.read())

            except Exception as error:
                msg = f'{type(error)}: {error}. src: "{mod.src}"'
                click.secho(msg, fg="bright_yellow")
                continue

            src_codes = OrderedDict()

            for el in src_mod.body:
                if not isinstance(el, (FunctionDef, ClassDef)):
                    continue

                qname = f"{mod.name}.{el.name.value}"

                if isinstance(exclude, str) and re.search(exclude, qname):
                    click.secho("excluding: ", nl=False, fg="bright_white")
                    click.secho(qname, fg="bright_black")

                    continue

                src_codes.update({qname: Code(module=mod, cst=el)})

                click.echo("\033[K", nl=False)
                click.secho("collecting: ", nl=False, fg="bright_white")
                click.secho(f"{qname}", fg="bright_cyan")

            if len(src_codes) == 0:
                continue

            try:
                with open(mod.dst, "r") as f:
                    test_mod = libcst.parse_module(f.read())

                # NOTE: let module's and class's body mutable
                test_mod = dataclasses.replace(test_mod, body=list(test_mod.body))
                for i, stmt in enumerate(test_mod.body):
                    if isinstance(stmt, ClassDef):
                        test_mod.body[i] = dataclasses.replace(
                            stmt,
                            body=dataclasses.replace(
                                stmt.body, body=list(stmt.body.body)
                            ),
                            # IndentedBlock(body=list(stmt.body.body))
                        )

                click.secho("loading: ", nl=False, fg="bright_white")
                click.secho(f"{mod.dst}", fg="bright_green")

            except FileNotFoundError:
                test_mod = Module(body=[])

            except Exception as error:
                msg = f'{type(error)}: {error}. dst: "{mod.dst}"'
                click.secho(msg, fg="bright_yellow")
                continue

            test_codes = OrderedDict()

            for i, el in enumerate(test_mod.body):
                if isinstance(el, (ClassDef, FunctionDef)):
                    test_pfx = (
                        Template.function_layout.prefix
                        if isinstance(el, FunctionDef)
                        else Template.class_layout.prefix
                    )

                    org_name = re.sub(rf"^{test_pfx}(.+)", r"\1", el.name.value)
                    key = f"{mod.name}.{org_name}"

                else:
                    key = f"#{i}"

                test_codes.update({key: Code(module=mod, cst=el)})

            for key, tcode in test_codes.items():
                if key not in src_codes:
                    continue

                scode = src_codes.pop(key)

                if not isinstance(scode.cst, ClassDef):
                    continue

                test_pfx = Template.class_layout.method_layout.prefix
                tmethods = OrderedDict()
                for j, el in enumerate(tcode.cst.body.body):
                    key = el.name.value if isinstance(el, FunctionDef) else f"#{j}"

                    tmethods.update({key: el})

                for el in scode.cst.body.body:
                    # TODO: allow private methods?
                    if (not isinstance(el, FunctionDef)) or el.name.value.startswith(
                        "__"
                    ):
                        continue

                    tname = f"{test_pfx}{el.name.value}"

                    if tname not in tmethods:
                        f: FunctionDef = Template.class_layout.method_layout(
                            Code(module=scode.module, cst=el)
                        ).build()

                        click.echo("\033[K", nl=False)
                        click.secho(
                            "adding method: ",
                            nl=False,
                            fg="bright_white",
                        )
                        click.secho(
                            f"{scode.cst.name.value}:{el.name.value}",
                            fg="bright_cyan",
                        )

                        tcode.cst.body.body.append(f)

                        code_added += 1

            for key, scode in src_codes.items():
                Layout = (
                    Template.function_layout
                    if isinstance(scode.cst, FunctionDef)
                    else Template.class_layout
                )

                what = "function" if isinstance(scode.cst, FunctionDef) else "class"

                obj = Layout(scode).build()
                click.echo("\033[K", nl=False)
                click.secho(f"adding {what}: ", nl=False, fg="bright_white")
                click.secho(f"{obj.name.value}", fg="bright_cyan")

                test_mod.body.append(obj)
                code_added += 1

            if code_added > 0:
                code = test_mod.code
                try:
                    source = blacking(code)
                except Exception as error:
                    msg = f'{type(error).__name__}: {error}. src: "{mod.src}"'
                    click.secho(msg, fg="bright_red")
                    click.secho(code, fg="bright_yellow")
                    continue

                if dryrun:
                    click.secho("(dryrun)", nl=False, fg="bright_yellow")
                    click.secho(f" writing: {mod.dst}", fg="bright_black")

                else:
                    makefile(mod.dst, source, overwrite=True)

                    click.secho("writing: ", nl=False, fg="bright_white")
                    click.secho(f"{mod.dst}", fg="bright_green")

            else:
                click.secho("all templates populated, skip.", fg="bright_black")

    populate_init(output)



import os
import re
from importlib.util import find_spec
from pathlib import Path
from typing import Generator, Optional, Union

from libcst import ClassDef, FunctionDef, Lambda

from .utils import Serializable, catch_module_from_sys, immutable


@immutable
class ModuleIO(Serializable):
    name: str
    outdir: str
    src: str
    dst: str

    @classmethod
    def from_name(
        cls, modname: str, outdir: str, head: str = None
    ) -> Optional["ModuleIO"]:
        with catch_module_from_sys(modname):
            try:
                spec = find_spec(modname)
            except Exception:
                return None

        if (spec is None) or (spec.origin is None) or (not os.path.isfile(spec.origin)):
            return None

        outdir = os.path.normpath(outdir)
        head = head or modname

        src = spec.origin
        if src.endswith("__init__.py"):
            pfx = modname
            base = "__init__"
        else:
            pfx, base = modname.rsplit(".", 1) if "." in modname else ("", modname)

        pfx = re.sub(rf"^{head}\.?", "", pfx)

        dst = os.path.join(
            outdir,
            head.replace(".", "_"),
            pfx.replace(".", os.path.sep),
            f"test_{base}.py",
        )

        return cls(name=modname, outdir=outdir, src=src, dst=dst)

    def iter_submodules(self, head: str = None) -> Generator["ModuleIO", None, None]:
        if self.ispkg:
            prefix = os.path.dirname(self.src)

            for p in Path(prefix).glob("**/*.py"):
                path = str(p)
                if path.endswith("__init__.py"):
                    path = os.path.dirname(path)

                suffix = re.sub(rf"^{prefix}", "", path)
                sub = os.path.splitext(suffix)[0]

                name = self.name + sub.replace(os.path.sep, ".")
                head = head or self.name
                mod = self.from_name(name, self.outdir, head=head)

                if mod:
                    yield mod

        else:
            yield self

    @property
    def ispkg(self):
        return self.src.endswith("__init__.py")


@immutable
class Code(Serializable):
    module: Optional[ModuleIO] = None
    cst: Union[FunctionDef, Lambda, ClassDef]



import dataclasses
from typing import List as LIST

from libcst import (
    Assert,
    BaseSuite,
    ClassDef,
    CSTNode,
    Decorator,
    FunctionDef,
    ImportAlias,
    ImportFrom,
    IndentedBlock,
    Name,
    Param,
    Parameters,
    Pass,
    SimpleStatementLine,
    parse_expression,
)

from .model import Code
from .utils import load_module_by_name


class Layout:
    """TODO: What is this"""

    prefix = "Test"
    ref = CSTNode

    def __init__(self, code: Code):
        self._code = code

    def build(self) -> CSTNode:
        fields = {f.name for f in dataclasses.fields(self.ref)}

        params = {fd: getattr(self, fd) for fd in fields if hasattr(self, fd)}

        return self.ref(**params)


class FunctionLayout(Layout):
    prefix = "test_"
    ref = FunctionDef

    @property
    def name(self) -> Name:
        return Name(value=f"{self.prefix}{self._code.cst.name.value}")

    @property
    def params(self) -> Parameters:
        return Parameters()

    @property
    def body(self) -> BaseSuite:
        return IndentedBlock(
            body=[
                SimpleStatementLine(
                    body=[
                        ImportFrom(
                            module=parse_expression(self._code.module.name),
                            names=[
                                ImportAlias(name=Name(value=self._code.cst.name.value))
                            ],
                        )
                    ]
                ),
                SimpleStatementLine(body=[Pass()]),
            ]
        )


class MethodLayout(FunctionLayout):
    @property
    def body(self) -> BaseSuite:
        return IndentedBlock(body=[SimpleStatementLine(body=[Pass()])])

    @property
    def params(self) -> Parameters:
        return Parameters(params=[Param(Name(value="self"))])


class ClassLayout(Layout):
    prefix = "Test"
    ref = ClassDef
    method_layout = MethodLayout

    @property
    def setups_teardowns(self) -> LIST[FunctionDef]:
        """TODO: what is this"""

        cls_params = Parameters(params=[Param(Name(value="cls"))])
        cls_deco = Decorator(decorator=Name(value="classmethod"))
        setup_class = FunctionDef(
            name=Name(value="setup_class"),
            body=IndentedBlock(
                body=[
                    SimpleStatementLine(
                        body=[
                            ImportFrom(
                                module=parse_expression(self._code.module.name),
                                names=[
                                    ImportAlias(
                                        name=Name(value=self._code.cst.name.value)
                                    )
                                ],
                            )
                        ]
                    ),
                    SimpleStatementLine(
                        body=[
                            Assert(test=Name(value=self._code.cst.name.value)),
                        ]
                    ),
                ]
            ),
            params=cls_params,
            decorators=[cls_deco],
        )

        teardown_class = FunctionDef(
            name=Name(value="teardown_class"),
            params=cls_params,
            body=IndentedBlock(body=[SimpleStatementLine(body=[Pass()])]),
            decorators=[cls_deco],
        )

        self_params = Parameters(
            params=[Param(Name(value="self")), Param(Name(value="method"))]
        )

        setup_method = FunctionDef(
            name=Name(value="setup_method"),
            params=self_params,
            body=IndentedBlock(body=[SimpleStatementLine(body=[Pass()])]),
        )

        teardown_method = FunctionDef(
            name=Name(value="teardown_method"),
            params=self_params,
            body=IndentedBlock(body=[SimpleStatementLine(body=[Pass()])]),
        )

        return [setup_class, teardown_class, setup_method, teardown_method]

    @property
    def methods(self) -> LIST[FunctionDef]:
        methods = []
        for stmt in self._code.cst.body.body:
            if (isinstance(stmt, FunctionDef)) and (
                not stmt.name.value.startswith("__")
            ):
                methods.append(self.method_layout(self._code.evolve(cst=stmt)).build())

        return methods

    @property
    def name(self) -> Name:
        return Name(value=f"{self.prefix}{self._code.cst.name.value}")

    @property
    def body(self) -> BaseSuite:
        return IndentedBlock(body=self.setups_teardowns + self.methods)


class Template:
    """TODO: what"""

    class_layout = ClassLayout
    function_layout = FunctionLayout

    @classmethod
    def load(cls, name: str) -> "Template":
        """TODO: what?"""

        modname, qname = name.rsplit(".", 1)

        mod = load_module_by_name(modname)

        return getattr(mod, qname)


class AssertFalseFunctionLayout(FunctionLayout):
    @property
    def body(self) -> BaseSuite:
        return IndentedBlock(
            body=[
                SimpleStatementLine(
                    body=[
                        ImportFrom(
                            module=parse_expression(self._code.module.name),
                            names=[
                                ImportAlias(name=Name(value=self._code.cst.name.value))
                            ],
                        )
                    ]
                ),
                SimpleStatementLine(
                    body=[
                        Assert(test=Name(value="False")),
                    ]
                ),
            ]
        )


class AssertSelfFunctionLayout(FunctionLayout):
    @property
    def body(self) -> BaseSuite:
        return IndentedBlock(
            body=[
                SimpleStatementLine(
                    body=[
                        ImportFrom(
                            module=parse_expression(self._code.module.name),
                            names=[
                                ImportAlias(name=Name(value=self._code.cst.name.value))
                            ],
                        )
                    ]
                ),
                SimpleStatementLine(
                    body=[
                        Assert(test=Name(value=self._code.cst.name.value)),
                    ]
                ),
            ]
        )


class AssertFalseMethodLayout(MethodLayout):
    @property
    def body(self):
        return IndentedBlock(
            body=[SimpleStatementLine(body[Assert(test=Name(value="False"))])]
        )


class AssertFalseClassLayout(ClassLayout):
    method_layout = AssertFalseMethodLayout


class AssertFalseTemplate(Template):
    function_layout = AssertFalseFunctionLayout
    class_layout = AssertFalseClassLayout


class AssertSelfTemplate(Template):
    function_layout = AssertSelfFunctionLayout


PassTemplate = Template



import json
import os
import pathlib
import pkgutil
import subprocess as sp
import sys
import tempfile
from contextlib import contextmanager
from importlib._bootstrap_external import SourceFileLoader
from types import ModuleType
from typing import Dict, Union

import attr
import cattr

immutable = attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)


@immutable
class Serializable:
    def to_dict(self, recurse=True, **kwargs):
        return attr.asdict(self, recurse=recurse, **kwargs)

    @classmethod
    def from_dict(cls, data: Dict):
        return cattr.structure_attrs_fromdict(data, cls)

    @classmethod
    def from_json(cls, json_str: str):
        return cls.from_dict(json.loads(json_str))

    def to_json(self, ensure_ascii=False, **kwargs):
        return json.dumps(self.to_dict(), ensure_ascii=ensure_ascii, **kwargs)

    def evolve(self, **kwargs):
        return attr.evolve(self, **kwargs)

    def mutate_from_other(self, other: "Serializable", excludes=[]):
        self_fields = [f2.name for f2 in attr.fields(self.__class__)]

        valid_fields = [
            f1.name
            for f1 in attr.fields(other.__class__)
            if (getattr(other, f1.name, None) is not None)
            and (f1.name not in excludes)
            and (f1.name in self_fields)
        ]

        data = other.to_dict(filter=lambda att, value: att.name in valid_fields)

        return self.evolve(**data)


def load_module_by_name(modname: str) -> ModuleType:
    loader: SourceFileLoader = pkgutil.get_loader(modname)
    return loader.load_module(modname)


def blacking(source_code: str):
    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        f.write(source_code)
        fname = f.name

    p = sp.Popen(["cat", fname], stdout=sp.PIPE)

    out = sp.check_output([sys.executable, "-m", "black", "-q", "-"], stdin=p.stdout)

    p.wait()

    try:
        pathlib.Path(fname).unlink()
    except FileNotFoundError:
        pass

    return out.decode()


def isorting(source_code: str):
    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        f.write(source_code)
        fname = f.name

    p = sp.Popen(["cat", fname], stdout=sp.PIPE)

    out = sp.check_output([sys.executable, "-m", "isort", "-q", "-"], stdin=p.stdout)

    p.wait()

    try:
        pathlib.Path(fname).unlink()
    except FileNotFoundError:
        pass

    return out.decode()


@contextmanager
def expand_sys_path(*paths: str):
    num = len(paths)

    for p in paths[::-1]:
        if p and isinstance(p, str):
            sys.path.insert(0, p)

    yield

    for _ in range(num):
        sys.path.pop(0)


def makefile(fullpath: str, content: Union[str, bytes] = "", overwrite: bool = False):
    if not os.path.isfile(fullpath):
        dirpath = os.path.dirname(fullpath)

        if not os.path.isdir(dirpath):
            os.makedirs(dirpath)

        _writefile(fullpath, content)

    elif overwrite:
        _writefile(fullpath, content)


def _writefile(fullpath: str, content: Union[str, bytes] = ""):
    if not isinstance(content, (str, bytes)):
        raise TypeError(f"content must be str or bytes, got: {type(content)}")

    with open(fullpath, "wb" if isinstance(content, bytes) else "w") as f:
        f.write(content)


@contextmanager
def catch_module_from_sys(modname: str):
    m = sys.modules.pop(modname, None)

    yield

    if m is not None:
        sys.modules[modname] = m


def populate_init(folder: str):
    root = pathlib.Path(folder)
    paths = {root}

    for path in root.glob("**/*.py"):
        if path.name == "__init__.py":
            continue

        parent = path.parent

        while parent > root:
            if not parent.joinpath("__init__.py").exists():
                paths.add(parent)

            parent = parent.parent

    for p in paths:
        makefile(os.path.join(str(p), "__init__.py"))
